/*
 * 68k.c: 68000 CPU emulation abstraction interface
 */

#include <stdio.h>
#include <stdlib.h>
#include "68k.h"
#include "genital.h"
#include "star/starcpu.h"
#include "vdp.h"

/* CPU functions */
int             (*CPURun)(int);
int             (*CPUInterrupt)(int);

/* internal handler prototypes */
unsigned        StarscreamReadRAMByte(unsigned);
unsigned        StarscreamReadRAMWord(unsigned);
void            StarscreamWriteRAMByte(unsigned, unsigned);
void            StarscreamWriteRAMWord(unsigned, unsigned);
unsigned        StarscreamFakeRead(unsigned);
void            StarscreamFakeWrite(unsigned, unsigned);
unsigned        TestRead1(unsigned);
unsigned        TestRead2(unsigned);

int             cpu_type = DEFAULT_CPU;         /* current CPU emulator */


/*
 * Starscream memory map (all NULL items initialized by CPUInit())
 */

struct STARSCREAM_PROGRAMREGION fetch_instructions[] =
{
        { 0x000000, 0x3fffff, (unsigned) NULL },
        { 0xff0000, 0xffffff, (unsigned) NULL },
        { 0xff000000, 0xff3fffff, (unsigned) NULL },
        { 0xfff00000, 0xffffffff, (unsigned) NULL },
        { -1, -1, NULL }
};
struct STARSCREAM_DATAREGION    read_data_byte[] =
{
        { 0x000000, 0x3fffff, NULL, (void *) NULL },
        { 0x400000, 0xa0ffff, StarscreamFakeRead, NULL },
        { 0xa10000, 0xa10fff, IOReadByte, NULL },
        { 0xa11000, 0xdfffff, StarscreamFakeRead, NULL },
        { 0xe00000, 0xffffff, StarscreamReadRAMByte, NULL },
        { -1, -1, NULL, NULL }
};
struct STARSCREAM_DATAREGION    read_data_word[] =
{
        { 0x000000, 0x3fffff, NULL, (void *) NULL },
        { 0x400000, 0xa0ffff, StarscreamFakeRead, NULL },
        { 0xa10000, 0xa10fff, IOReadWord, NULL },
        { 0xa11000, 0xbfffff, StarscreamFakeRead, NULL },
        { 0xc00000, 0xdfffff, VDPReadWord, NULL },
        { 0xe00000, 0xffffff, StarscreamReadRAMWord, NULL },
        { -1, -1, NULL, NULL }
};
struct STARSCREAM_DATAREGION    write_data_byte[] =
{
        { 0x000000, 0xa0ffff, StarscreamFakeWrite, NULL },
        { 0xa10000, 0xa10fff, IOWriteByte, NULL },
        { 0xa11000, 0xbfffff, StarscreamFakeWrite, NULL },
        { 0xc00000, 0xdfffff, VDPWriteByte, NULL },
        { 0xe00000, 0xffffff, StarscreamWriteRAMByte, NULL },
        { -1, -1, NULL, NULL }
};
struct STARSCREAM_DATAREGION    write_data_word[] =
{
        { 0x000000, 0xa0ffff, StarscreamFakeWrite, NULL },
        { 0xa10000, 0xa10fff, IOWriteWord, NULL },
        { 0xa11000, 0xbfffff, StarscreamFakeWrite, NULL },
        { 0xc00000, 0xdfffff, VDPWriteWord, NULL }, 
        { 0xe00000, 0xffffff, StarscreamWriteRAMWord, NULL },
        { -1, -1, NULL, NULL }
};


/*
 * Voltron memory map
 */

/* struct VOLTRON_DATAREGIONREAD   voltron_read_byte[] =
{
        { 0x000000, 0x3fffff, TestRead1 },
        { 0x400000, 0xffffff, TestRead2 },
        { -1, -1, NULL }
}; */


/*
 * Starscream specific function handlers
 */

unsigned StarscreamReadRAMByte(unsigned address)
{
        return genesis.ram[(address ^ 1) & 0xffff];
}
        
unsigned StarscreamReadRAMWord(unsigned address)
{
        return *((unsigned short *) (genesis.ram + (address & 0xfffe)));
}

void StarscreamWriteRAMByte(unsigned address, unsigned data)
{
        genesis.ram[(address ^ 1) & 0xffff] = data;
}

void StarscreamWriteRAMWord(unsigned address, unsigned data)
{
        *((unsigned short *) (genesis.ram + (address & 0xfffe))) = data;
}

unsigned StarscreamFakeRead(unsigned address)
{
        return 0;
}

void StarscreamFakeWrite(unsigned address, unsigned data)
{       
}

int StarscreamRun(int cycles)
{
        unsigned        status;

        s68000tripOdometer();           /* reset odometer */

        status = s68000exec(cycles);
        if (status != 0x80000000)       /* an error occured! */
        {
                if (status == 0x80000001)
                        printf("Genital: Starscream: Out of bounds" \
                                " (PC = %06X)\n", s68000readPC());
                else if (status == 0x80000002)
                        printf("Genital: Starscream: Unsupported "  \
                                "stack frame\n");
                else if (status == 0xffffffff)
                        printf("Genital: Starscream: Double fault\n");
                else
                        printf("Genital: Invalid instruction at %06X\n",
                               status);
                return 1;
        }

        return 0;
}

int StarscreamInterrupt(int int_type)
{
        if (int_type == VINTERRUPT)
                if (s68000interrupt(6, -1) == 0)
                        return 1;
        return 0;
}
        
void StarscreamResetHandler()
{
}


/*
 * Voltron functions
 */

unsigned TestRead1(unsigned addr)
{
        printf("Called TestRead1 with addr = %08X\n", addr);
        return 0;
}
unsigned TestRead2(unsigned addr)
{
        printf("Called TestRead2 with addr = %08X\n", addr);
        return 0;
}



/*
 *  General 68000 functions
 */

unsigned CPUGetPC()
{
        if (cpu_type == STARSCREAM_CPU)
                return s68000readPC();
        return 0;
}

void CPUSetPC(unsigned data)
{
        if (cpu_type == STARSCREAM_CPU)
                s68000context.pc = data;
}

unsigned CPUGetSR()
{
        if (cpu_type == STARSCREAM_CPU)
                return s68000context.sr;
        return 0;
}

void CPUSetSR(unsigned data)
{
        if (cpu_type == STARSCREAM_CPU)
                s68000context.sr = data;
}

unsigned CPUGetA(int i)
{
        i &= 7;
        if (cpu_type == STARSCREAM_CPU)
                return s68000context.areg[i];
        return 0;
}

void CPUSetA(int i, unsigned data)
{
        i &= 7;
        if (cpu_type == STARSCREAM_CPU)
                s68000context.areg[i] = data;
}        

unsigned CPUGetD(int i)
{
        i &= 7;
        if (cpu_type == STARSCREAM_CPU)
                return s68000context.dreg[i];
        return 0;
}

void CPUSetD(int i, unsigned data)
{
        i &= 7;
        if (cpu_type == STARSCREAM_CPU)
                s68000context.dreg[i] = data;
}        


void CPUReset()
{
        if (cpu_type == STARSCREAM_CPU)
                s68000reset();
}

void CPUInit(int type)
{
        if (type == STARSCREAM_CPU)     /* Starscream core */
        {
                cpu_type = STARSCREAM_CPU;

                /* set some context pointers */
                fetch_instructions[0].offset
                                = (unsigned) genesis.rom - 0x000000;
                fetch_instructions[1].offset
                                = (unsigned) genesis.ram - 0xff0000;
                fetch_instructions[2].offset
                                = (unsigned) genesis.rom - 0xff000000; //c
                fetch_instructions[3].offset
                                = (unsigned) genesis.ram - 0xfff00000;
                read_data_byte[0].userdata = (void *) genesis.rom;
                read_data_word[0].userdata = (void *) genesis.rom;

                /* initialize the default context provided by Starscream */
                s68000context.fetch = fetch_instructions;
                s68000context.readbyte = read_data_byte;
                s68000context.readword = read_data_word;
                s68000context.writebyte = write_data_byte;
                s68000context.writeword = write_data_word;
                s68000context.s_fetch = fetch_instructions;
                s68000context.s_readbyte = read_data_byte;
                s68000context.s_readword = read_data_word;
                s68000context.s_writebyte = write_data_byte;
                s68000context.s_writeword = write_data_word;
                s68000context.u_fetch = fetch_instructions;
                s68000context.u_readbyte = read_data_byte;
                s68000context.u_readword = read_data_word;
                s68000context.u_writebyte = write_data_byte;
                s68000context.u_writeword = write_data_word;
                s68000context.resethandler = StarscreamResetHandler;

                /* set up handlers */
                CPURun = StarscreamRun;
                CPUInterrupt = StarscreamInterrupt;

                /* Starscream requires bytes to be swapped */
                SwapBuffer(genesis.rom, 4194304);

                /* initialize Starscream */
                s68000init();
                if (s68000reset())
                {
                        printf("Genital: Starscream failed to initialize\n");
                        exit(1);
                }
        }
        else
        {
                printf("Genital: Unknown CPU emulator type: %d\n", type);
                exit(1);
        }
}

