/*                                                     
 * vdp.c: VDP emulation
 */

#include <stdio.h>
#include <stdlib.h>
#include "68k.h"
#include "misc.h"
#include "genital.h"
#include "vdp.h"
#include "vga.h"

struct vdp_struct       vdp;


/*
 * Sprite order
 */

void VDPCalcSpriteOrder()
{
        int             i;
        unsigned        next = 0;

        for (i = 0; i < 80; i++)
        {
                vdp.sprite_order[i] = next;
                /* ^1 for little endian */
                next = vdp.sprite_addr[((next << 3) + 3) ^ 1];
                if (next == 0)
                {
                        vdp.sprite_count = i + 1;       /* # of sprites */
                        return;
                }
        }
}

/*
 * VDP RAM access
 */

inline void WriteVDPRAMByte(unsigned data)
{
        /* Fix: This is probably wrong... */           

        if (vdp.access_mode == 1)       /* access mode 1 = VRAM write */
        {
                if (vdp.addr & 1)
                        /* if odd, simply write the byte */
                        vdp.vram[vdp.addr & 0xfffe] = data;
                else
                        /* otherwise, write it at next position */
                        vdp.vram[vdp.addr + 1] = data;

                vdp.addr += vdp.reg[15];        /* auto increment */
        }
}
        
inline void WriteVDPRAMWord(unsigned data)
{
        /* write word to VDP RAM (uses VDP context data for address and which
           RAM to write) */
        if (vdp.access_mode == 1)       /* access mode 1 = VRAM write */
        {
                if (vdp.addr & 1)
                        WORDSWAP(data);
                PutWord(&vdp.vram[vdp.addr & 0xfffe], data);
                vdp.addr += vdp.reg[15];        /* auto increment */
        }
        else if (vdp.access_mode == 3)  /* mode 3 = CRAM write */
        {
                PutWord(&vdp.cram[vdp.addr], data);
                SetPalette(vdp.addr, data);
                vdp.addr += vdp.reg[15];
        }
        else if (vdp.access_mode == 5)  /* mode 5 = VSRAM write */
        {
                PutWord(&vdp.vsram[vdp.addr & 0x7e], data);
                vdp.addr += vdp.reg[15];
        }
}        


/*
 * DMA (much is taken from DGen/SDL)
 */

void DoDMA()
{
        unsigned        i, len, addr;
        unsigned short  data;

        /* source address */
        addr = (vdp.reg[23] & 0x7f) << 17;
        addr += vdp.reg[22] << 9;
        addr += vdp.reg[21] << 1;
//        printf("source addr: %06X\n", addr);
//        printf("dest addr: %04X\n", vdp.addr);

        len = (vdp.reg[20] << 8) + vdp.reg[19]; /* length */
//        printf("dma len: %d\n", len);

        vdp.dma_mode = (vdp.reg[23] >> 6) & 3;
//        printf("dma mode: %d\n", vdp.dma_mode);

        switch (vdp.dma_mode)
        {                
                case 0: case 1:                 /* DMD1 = 0, DMD0 = 0 or 1,
                                                   memory-to-VRAM */                        
                        for (i = 0; i < len; i++)
                        {
                                data = MemReadWord(addr);       /* copy */
                                addr += 2;
                                WriteVDPRAMWord(data);
                        }
                        break;
                case 2:                         /* VRAM fill */
                        /* this is done on a write to the data port */
                        break;
                default:
                        break;
        }
}

void DoDMAFill(unsigned data, int word) /* word = 1: use words, else bytes */
{
        int     i, len;

        len = (vdp.reg[20] << 8) + vdp.reg[19]; /* length */

        for (i = 0; i < len; i++)
//                if (word)
                        WriteVDPRAMWord(data);
//                else
//                        WriteVDPRAMByte(data);

        vdp.dma_mode = 0;
}


/*
 * Registers
 */

void WriteReg(int i, unsigned data)
{
        vdp.reg[i] = data & 0xff;
}

void ScrollAAddr(int i, unsigned data)
{
        vdp.scroll_a_addr = &vdp.vram[(data & 0x38) << 10];        
}

void ScrollBAddr(int i, unsigned data)
{
        vdp.scroll_b_addr = &vdp.vram[(data & 7) << 13];
}

void SpriteAddr(int i, unsigned data)
{
        vdp.sprite_addr = &vdp.vram[(data & 0x7f) << 9];
}


void HScrollAddr(int i, unsigned data)
{        
        vdp.hscroll_addr = &vdp.vram[(data & 0x3f) << 10];
//        printf("hscroll: %04X\n", (data & 0x3f) << 10);
}


/*
 * Ports
 */

void WriteDataPortByte(unsigned data)
{
        if (vdp.dma_mode == 2)  /* do VRAM fill */
                DoDMAFill(data, 0);
        else
                WriteVDPRAMByte(data);
}

void WriteDataPortWord(unsigned data)
{
        if (vdp.dma_mode == 2)  /* do VRAM fill */
                DoDMAFill(data, 1);
        else 
                WriteVDPRAMWord(data);
}

void DecodeCtrlPortCommand()
{
        /* access mode consists of cd0-cd3, cd4-cd5 are not needed */
        vdp.access_mode = (vdp.control_port >> 30) +
                          ((vdp.control_port >> 2) & 0xc);
        /* get address */
        switch (vdp.access_mode)
        {
                case 1:         /* VRAM write */
                        vdp.addr = (vdp.control_port >> 16) & 0x3fff;
                        vdp.addr += (vdp.control_port & 3) << 14;
                        break;
                case 3:         /* CRAM write */
                        vdp.addr = (vdp.control_port >> 16) & 0x7f;
                        break;
                case 5:         /* VSRAM write */
                        vdp.addr = (vdp.control_port >> 16) & 0x7f;
                        break;
                default:
                        break;
        }

        if (vdp.control_port & 0x80)    /* cd5 = 1 for DMA */
                DoDMA();
}

void WriteCtrlPortWord(unsigned data)
{
        typedef         void (*reg_call)(int, unsigned);
        reg_call        Reg[24] = { WriteReg, WriteReg, ScrollAAddr, WriteReg,
                                    ScrollBAddr, SpriteAddr,
                                    WriteReg, WriteReg, WriteReg, WriteReg,
                                    WriteReg, WriteReg, WriteReg, HScrollAddr,
                                    WriteReg, WriteReg, WriteReg, WriteReg,
                                    WriteReg, WriteReg, WriteReg, WriteReg,
                                    WriteReg, WriteReg };

        if ((data >> 14) == 2)          /* write1 mode */
        {
                if (((data >> 8) & 0x1f) < 24)  /* make sure it is in range */
                        Reg[(data >> 8) & 0x1f]((data >> 8) & 0x1f, data);
        }
        else
        {
                if (vdp.ctrlport_mode == 0)     /* first word */
                {
                        vdp.control_port = data << 16;
                        vdp.ctrlport_mode++;    /* process word 2 next time */
                }
                else                            /* second word */
                {
                        vdp.control_port += data;
                        vdp.ctrlport_mode = 0;
                        DecodeCtrlPortCommand();        /* decode ctrl data */
                }
        }
}       


/*
 * Read/write handlers
 */

void VDPWriteByte(unsigned addr, unsigned data)
{
        data &= 0xff;
       
        /* lower byte to data port */
        if (addr == 0xc00001 || addr == 0xc00003)
                WriteDataPortByte(data);
}

void VDPWriteWord(unsigned addr, unsigned data)
{        
        data &= 0xffff;

        if (addr == 0xc00000 || addr == 0xc00002)       /* data port */
                WriteDataPortWord(data);
        else if (addr == 0xc00004 || addr == 0xc00006)  /* control port */
                WriteCtrlPortWord(data);
}

unsigned VDPReadWord(unsigned addr)
{
        if (addr == 0xc00004 || addr == 0xc00006)       /* status register */
        {
                vdp.status_register ^= 0x10;            /* FIFO kludge */
                vdp.status_register ^= 0x20;
                return vdp.status_register & 0xffff;
        }
        if (addr == 0xc00008 || addr == 0xc0000a)       /* HV counter */
                return vdp.hv_counter & 0xffff;

        return 0;
}


/*
 * HV counter management
 */

void VDPUpdateHVCounter(unsigned line)
{
        if (!(vdp.reg[0] & 2))  /* if counter is stopped, don't update */
                return;

        if (vdp.reg[12] & 1)    /* if h40 resolution... */
                vdp.hv_counter = (line << 8) | 0x9f;
        else                    /* h32 resolution... */
                vdp.hv_counter = (line << 8) | 0x7f;
}


/*
 * Interrupt management
 */

void VDPVBlankStart()
{                      
        if (vdp.reg[1] & 0x20)          /* if vertical interrupt allowed */
                CPUInterrupt(VINTERRUPT);
        vdp.status_register |= 0x80;    /* "V interrupt happened" */
        vdp.status_register |= 4;       /* in VBlank */
}

void VDPVBlankEnd()
{
        vdp.status_register ^= 0x80;    /* done with V interrupt */
        vdp.status_register ^= 4;       /* done w/ VBlank */
}               


/*
 * Initialization
 */

void VDPInit()
{
        if ((vdp.vram = calloc(65536, sizeof(unsigned char))) == NULL)
        {
                fprintf(stderr, "Genital: Failed to allocate 64KB RAM for " \
                        "VRAM\n");
                exit(1);
        }

        if ((vdp.cram = calloc(128, sizeof(unsigned char))) == NULL)
        {
                fprintf(stderr, "Genital: Failed to allocate 128 bytes of " \
                        "RAM for CRAM\n");
                exit(1);
        }

        if ((vdp.vsram = calloc(80, sizeof(unsigned char))) == NULL)
        {
                fprintf(stderr, "Genital: Failed to allocate 80 bytes of " \
                        "RAM for VSRAM\n");
                exit(1);
        }

        /* all register values are undefined at start up, pointers are
           initialized to prevent faults */

        vdp.scroll_a_addr = vdp.scroll_b_addr = vdp.hscroll_addr = vdp.vram;
        vdp.ctrlport_mode = vdp.dma_mode = vdp.hv_counter = 0;
        vdp.status_register = 0x200;
}

void VDPDeinit()
{
        free(vdp.vram);
        free(vdp.cram);
}
